library nui_theme;

import 'dart:convert';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/model/interfaces.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_theme/nui_color_theme.dart';
import 'package:nui_theme/nui_text_theme.dart';
import 'package:nui_theme/nui_theme_config.dart';
import 'package:nui_web/nui_web.dart';

export 'package:nui_theme/nui_color_theme.dart';
export 'package:nui_theme/nui_text_theme.dart';
export 'package:nui_theme/nui_theme_config.dart';
export 'package:nui_theme/nui_theme_widget.dart';

NUIColorTheme get colorTheme {
  return NUITheme.get().theme;
}
NUITextTheme get textTheme {
  return NUITheme.get().theme.textTheme;
}

class NUITheme{
  static NUITheme _instance;

  static const String MODULE = "NUITheme";
  static const String _PREF_CURRENT_THEME = "NUIThemeCurrentTheme";
  static const String _PREF_THEME_CACHE = "NUIThemeCache";
  static const String _SELECTION_DARK = "dark";
  static const String _SELECTION_LIGHT = "light";
  static const String BROADCAST_THEME_MODULE = "NUIThemeBroadcast";
  static const String CODE_THEME_SWITCHED = "Theme Switched";
  static const String CODE_THEME_UPDATED = "Theme Updated";

  String _urlPath;
  String _assetPath;
  bool _backgroundLoad;
  NUIThemeConfig _theme;

  // static NUITheme get instance => _instance;

  static NUITheme get(){
    if(_instance == null){
      throwNUI("NUITheme", "You have not initialized NUITheme with the builder");
      return null;
    }
    return _instance;
  }

  ThemeData getThemeData(){
    final isLightTheme = !isDarkTheme();
    return ThemeData(
        brightness: isLightTheme != false ? Brightness.light : Brightness.dark,
        primaryColor: theme.primary(),
        canvasColor: theme.card(),
        backgroundColor: theme.background(),
        accentColor: theme.primary(),
        primarySwatch: theme.primaryVariant(),
        textTheme: theme.textTheme.getTextTheme(),
        fontFamily: theme.textTheme.fontFamily,
        colorScheme: theme.getColorScheme()
    );
  }

  NUITheme._builder(NUIThemeBuilder builder){
    _urlPath = builder._urlPath;
    _assetPath = builder._assetPath;
    _backgroundLoad = builder._backgroundLoad ?? true;
    _instance = this;
  }

  Future<DataResult> _initializeTheme() async{
    DataResult loadResult;
    if(!isNullOrEmpty(_assetPath)){
      loadResult = await _loadThemeFromAsset();
    }
    else if(!isNullOrEmpty(_urlPath)){
      loadResult = await _loadThemeFromNetwork();
    }
    else{
      //Both asset path and url path is null, this should never happen
      loadResult = DataResult(success: false, message: "Both asset and network paths are null");
      throwNUI(MODULE, "Please specify either the asset path or the network path to fetch the theme configuration");
    }
    await _initializePreference();
    await _initializeBusEvent();
    return loadResult;
  }

  Future<bool> _saveThemeToCache(NUIThemeConfig theme) async{
    final result = await NUISharedPref.get().save(_PREF_THEME_CACHE, theme);
    return true;
  }

  Future<bool> _loadThemeFromCache() async{
    final result = await NUISharedPref.get().getObject(_PREF_THEME_CACHE, NUIThemeConfigEnt());
    if(result != null){
      _theme = result;
      return true;
    }
    return false;
  }

  Future<bool> _updateSelectedTheme(bool isDark) async{
    if(isDark){
      await NUISharedPref.get().save(_PREF_CURRENT_THEME, _SELECTION_DARK);
    }
    else{
      await NUISharedPref.get().save(_PREF_CURRENT_THEME, _SELECTION_LIGHT);
    }
    return true;
  }

  NUIColorTheme get theme {
    final isDark = isDarkTheme();
    if(isDark){
      _theme.darkTheme.setColorCodes(_theme.colorCodes);
      return _theme.darkTheme;
    }
    _theme.lightTheme.setColorCodes(_theme.colorCodes);
    return _theme.lightTheme;
  }

  Future<bool> switchTheme() async{
    //No dark theme in the configuration, so skip this
    if(_theme.darkTheme == null) return false;

    final isDark = isDarkTheme();
    if(isDark){
      logNUI("NUITheme", "Switching from dark theme to light theme");
      await _updateSelectedTheme(false);
    }
    else{
      logNUI("NUITheme", "Switching from light theme to dark theme");
      await _updateSelectedTheme(true);
    }
    NUIBusEvent.get(module: BROADCAST_THEME_MODULE).send(NUIBusEventData<NUIColorTheme>(data: theme, sender: MODULE, code: CODE_THEME_SWITCHED));
    return true;
  }

  String _getSelectedTheme(){
    final selectedTheme = NUISharedPref.get().quickGetString(_PREF_CURRENT_THEME, defValue: "");
    return selectedTheme;
  }

  bool isDarkTheme(){
    final selectedTheme = _getSelectedTheme();
    if(match(selectedTheme, _SELECTION_DARK)){
      return true;
    }
    return false;
  }

  Future<bool> _initializePreference() async{
    final sharedPref = await NUISharedPrefBuilder().build();
    return sharedPref != null;
  }

  Future<bool> _initializeBusEvent() async{
    final busEvent = NUIBusEventBuilder()
        .module(BROADCAST_THEME_MODULE)
        .build();
    return true;
  }

  Future<DataResult> _loadThemeFromAsset() async{
    try{
      logNUI(MODULE, "Loading theme configuration from asset path : $_assetPath");
      final json = await rootBundle.loadString(_assetPath);
      final theme = NUIThemeConfigEnt().toEntity(jsonDecode(json));
      _theme = theme;
      return DataResult(success: true, message: "Success");
    }catch(e){
      logNUI(MODULE, "Failed to load theme configuration from asset path with error : $e");
      return DataResult(success: false, message: "Failed to load theme configuration from asset path : $_assetPath");
    }
  }

  Future<DataResult> _loadThemeFromNetwork() async{
    await _loadThemeFromCache();

    final futureResult = _callApiToFetchTheme();
    DataResult result;

    //Background load is turned off or no previous theme was found
    if(_backgroundLoad != true || _theme == null) {
      result = await futureResult;
    }
    else{
      result = DataResult(success: true, message: "Proceeding with cache theme");
    }

    return result;
  }

  Future<DataResult<NUIThemeConfig>> _callApiToFetchTheme() async{
    final web = await NUIWebBuilder()
        .baseURL(_urlPath)
        .module(MODULE)
        .build();

    final futureResult = web.doGET(entity: NUIThemeConfigEnt());

    final webResult = await futureResult;
    if (webResult.success && webResult.body != null) {
      _theme = webResult.body;
      await _saveThemeToCache(_theme);
      NUIBusEvent.get(module: BROADCAST_THEME_MODULE).send(NUIBusEventData<NUIColorTheme>(data: theme, sender: MODULE, code: CODE_THEME_UPDATED));
      return DataResult(success: true, message: "Success");
    }
    else {
      return DataResult(success: false, message: "Failed to load them configuration from network path : $_urlPath");
    }
  }

}

class NUIThemeBuilder{
  String _assetPath;
  String _urlPath;
  bool _backgroundLoad;

  NUIThemeBuilder fromAsset(String assetPath){
    _assetPath = assetPath;
    return this;
  }

  NUIThemeBuilder fromNetwork(String url){
    _urlPath = url;
    return this;
  }

  NUIThemeBuilder backgroundLoad(bool backgroundLoad){
    _backgroundLoad = backgroundLoad;
    return this;
  }

  Future<NUITheme> build() async{
    final theme = NUITheme._builder(this);
    final loadResult = await theme._initializeTheme();
    return theme;
  }
}